--------------------------- MODULE Determinism -------------------------------
EXTENDS System_Spec, Filters, Witness

(***************************************************************************)
(* This module contains the proof of the main determinism theorem: for any *)
(* fixed witness execution of the system, the sequence of states any agent *)
(* goes through in the actual execution agrees with the sequence in the    *)
(* witness execution.                                                      *)
(***************************************************************************)

USE DEF Time, Natp

Max(m,n) == IF m >= n THEN m ELSE n 
USE DEF Max


(***************************************************************************)
(* The following formula asserts that the witness contains all information *)
(* about the current history of every agent.                               *)
(***************************************************************************)
Det(witness) == 
  \E U \in Natp : \A a \in Agent :
     IsPrefix(filter(history, a), filter(Restrict(witness, 1..U), a))

THEOREM Determinism == 
  ASSUME NEW witness \in Stream(GoodSystemState), Is_witness(witness)
  PROVE  Spec => [] Det(witness)
<1> witness \in Stream(SystemState)
    BY DEF GoodSystemState, Stream
<1>1. Init /\ History_type => Det(witness)
  <2>.  HAVE Init /\ History_type
  <2>1. Restrict(witness, 1..1) = <<witness[1]>>
    <3>. witness[1] \in GoodSystemState
      BY DEF Stream
    <3>. /\ Restrict(witness, 1..1) \in Seq(GoodSystemState)
         /\ <<witness[1]>> \in Seq(GoodSystemState)
         /\ Len(Restrict(witness, 1..1)) = 1
         /\ Len(<<witness[1]>>) = 1
      BY RestrictStream, Isa
    <3>. \A i \in 1..1 : Restrict(witness, 1..1)[i] = <<witness[1]>>[i]
      BY DEF Restrict, Stream
    <3>. QED  BY SeqEqual, Zenon
  <2>2. Restrict(witness, 1..1) = history
    BY <2>1 DEF Init, Is_witness
  <2>. DEFINE fwit(a) == filter(Restrict(witness, 1..1), a)
  <2>3. \A a \in Agent : IsPrefix(filter(history, a), fwit(a))
    BY <2>2, IsPrefixPartialOrder, filter_type DEF History_type
  <2>. QED  BY <2>3, Isa DEF Det
<1>2. /\ Det(witness)
      /\ [Next]_vars
      /\ History_type /\ History_type'
      /\ State_type /\ State_type'
      /\ Inv
      /\ Inv_messages /\ Inv_messages'
      => Det(witness)'
  <2>. SUFFICES ASSUME Det(witness),
                       [Next]_vars,
                       History_type, History_type',
                       State_type, State_type',
                       Inv, Inv_messages, Inv_messages'
                PROVE  Det(witness)'
    OBVIOUS
  <2>1. ASSUME time' \in Time, time' >= time,
               \A b \in Agent : time' <= instr(state[b]).late,
               NEW a \in Agent,
               instr(state[a]).early <= time',
               LET res == exec(state[a], messages)
               IN  /\ state' = [state EXCEPT ![a] = res[1]]
                   /\ messages' = res[2],
               history' = Append(history, [st |-> state', msgs |-> messages', t |-> time'])
        PROVE  Det(witness)'
    <3>1. PICK U \in Natp : 
               \A b \in Agent : IsPrefix(filter(history, b), filter(Restrict(witness, 1..U), b))
      BY DEF Det
    <3>. /\ Restrict(witness, 1..U) \in Seq(GoodSystemState)
         /\ Len(Restrict(witness, 1..U)) = U
      BY RestrictStream DEF Stream
    <3>2. PICK N \in Natp : /\ filter(history, a) = filter(Restrict(witness, 1..N), a)
                            /\ executes(a, witness[N], witness[N+1])
      <4>1. PICK w \in Seq(SystemState) :
                 /\ IsPrefix(w, Restrict(witness, 1..U))
                 /\ filter(w,a) = filter(history,a)
        BY IsPrefix_filter, <3>1 DEF History_type, GoodSystemState
      <4>2. PICK M \in 0 .. U : w = Restrict(witness, 1..M)
        BY <4>1, IsPrefixRestrictStream DEF Stream, GoodSystemState
      <4>3. w # << >>
        BY <4>1, filter_empty DEF History_type
      <4>4. M # 0
        BY <4>2, <4>3, RestrictStream
      <4>5. PICK N \in Natp : /\ N >= M
                              /\ \A k \in M .. N : witness[k].st[a] = witness[M].st[a]
                              /\ executes(a, witness[N], witness[N+1])
        BY witness_executes, <4>4, Z3 DEF Stream
      <4>6. filter(Restrict(witness, 1..N), a) = filter(Restrict(witness, 1..M), a)
        BY witness_filter_extend, <4>4, <4>5 DEF Stream, GoodSystemState
      <4>. QED  BY <4>1, <4>2, <4>5, <4>6

    <3>. DEFINE newU == Max(U, N+1)
    <3>3. /\ newU \in Natp
          /\ U <= newU
          /\ N <= newU
          /\ N+1 <= newU
      OBVIOUS
    <3>. HIDE DEF newU
    <3>. [st |-> state', msgs |-> messages', t |-> time'] \in SystemState
      BY DEF State_type, SystemState
    <3>4. /\ Restrict(witness, 1..N) \in Seq(SystemState)
          /\ Restrict(witness, 1..N+1) \in Seq(SystemState)
          /\ Restrict(witness, 1..newU) \in Seq(SystemState)
          /\ IsPrefix(Restrict(witness, 1..U), Restrict(witness, 1 .. newU))
          /\ IsPrefix(Restrict(witness, 1 .. N+1), Restrict(witness, 1 .. newU))
      BY <3>3, RestrictStream, RestrictStreamMonotonic DEF GoodSystemState
    <3>5. \A b \in Agent :
             /\ filter(history, b) \in Seq(State)
             /\ filter(history', b) \in Seq(State)
             /\ filter(Restrict(witness, 1 .. U), b) \in Seq(State)
             /\ filter(Restrict(witness, 1 .. newU), b) \in Seq(State)
             /\ IsPrefix( filter(Restrict(witness, 1 .. U), b),
                          filter(Restrict(witness, 1 .. newU), b))
             /\ IsPrefix( filter(Restrict(witness, 1 .. N+1), b),
                          filter(Restrict(witness, 1 .. newU), b))
      BY <3>4, filter_type, filter_IsPrefix DEF History_type, GoodSystemState
    <3>6. ASSUME NEW b \in Agent
          PROVE  IsPrefix(filter(history', b), filter(Restrict(witness, 1 .. newU), b))
      <4>1. CASE b = a
        <5>. SUFFICES filter(history', a) = filter(Restrict(witness, 1 .. N+1), a)
          BY <4>1, <3>5, IsPrefixPartialOrder
        <5>. DEFINE wit_st == witness[N].st[a]
                    wit_ins == instr(wit_st)
        <5>1. wit_st = state[a]
          <6>1. witness[N] = Last(Restrict(witness, 1..N))
            BY DEF Restrict, Last
          <6>2. DOMAIN Restrict(witness, 1..N) # DOMAIN <<>>
            BY DEF Restrict
          <6>3. Restrict(witness, 1..N) # <<>>
            BY <6>2
          <6>4. wit_st = Last(filter(Restrict(witness, 1..N), a))
            BY <3>4, <6>1, <6>2, Last_filter
          <6>5. Last(filter(history, a)) = state[a]
            BY Last_filter DEF History_type, Inv
          <6>. QED  BY <6>4, <6>5, <3>2
        <5>2. ASSUME wit_ins.ins.type = "receive"
              PROVE  msgReceived(witness[N].msgs, wit_ins) <=> msgReceived(messages, wit_ins)
          <6>1. ASSUME msgReceived(witness[N].msgs, wit_ins)
                PROVE  msgReceived(messages, wit_ins)
            <7>1. PICK msg \in witness[N].msgs : /\ wit_ins.ins.msg = msg.ins.msg
                                                 /\ visible(msg, wit_ins)
                                                 /\ msg.late < wit_ins.early
              BY <6>1, visible_cond, instr_type DEF msgReceived, SystemState, Message, Stream, GoodSystemState
            <7>msg. msg \in Message
              BY <7>1 DEF Stream, GoodSystemState, SystemState
            <7>2. PICK c \in Agent : msg \in sent(Range(project_agent(Restrict(witness, 1..N),c)))
              BY <7>1, witness_messages DEF Stream, GoodSystemState
            <7>3. PICK K \in 1..N : msg = instr(witness[K].st[c]) /\ msg.ins.type = "send"
              BY <7>2 DEF sent, Range, project_agent, Restrict, Message
            <7>4. PICK M \in 1..U : filter(history, c) = filter(Restrict(witness, 1..M), c)
              <8>1. PICK w \in Seq(SystemState) : 
                         /\ IsPrefix(w, Restrict(witness, 1..U))
                         /\ filter(w,c) = filter(history,c)
                BY IsPrefix_filter, <3>1, <3>4 DEF History_type, GoodSystemState
              <8>2. PICK M \in 0 .. U : w = Restrict(witness, 1..M)
                BY <8>1, IsPrefixRestrictStream
              <8>3. w # << >>
                BY <8>1, filter_empty DEF History_type
              <8>4. M # 0
                BY <8>2, <8>3, RestrictStream
              <8>. QED  BY <8>1, <8>2, <8>4
            <7>5. /\ Restrict(witness, 1..M) \in Seq(SystemState)
                  /\ Len(Restrict(witness, 1..M)) = M
                  /\ Restrict(witness, 1..M) # << >>
              BY RestrictStream
            <7>6. witness[M].st[c] = state[c]
              <8>1. Last(Restrict(witness, 1..M)) = witness[M]
                BY <7>5 DEF Restrict, Last
              <8>2. Last(history).st[c] = state[c]
                BY DEF Inv
              <8>. QED  BY <7>4, <7>5, <8>1, <8>2, Last_filter DEF History_type
            <7>. /\ wit_ins.early \in Time
                 /\ time' \in Time
                 /\ msg.late \in Time
                 /\ instr(witness[M].st[c]).late \in Time
              BY instr_type DEF State_type, DatedInstruction, SystemState, Message, Stream
            <7>7. wit_ins.early <= instr(witness[M].st[c]).late
              BY <2>1, <5>1, <7>6, Z3
            <7>8. K < M
              <8>. SUFFICES ASSUME M <= K PROVE FALSE
                OBVIOUS
              <8>. instr(witness[M].st[c]).late <= instr(witness[K].st[c]).late
                BY witness_monotonic_deadlines DEF Stream
              <8>. QED
                BY <7>1, <7>3, <7>7, Z3
            <7>9. PICK j \in 1 .. Len(history) : history[j].st[c] = witness[K].st[c]
              <8>1. witness[K].st[c] \in Range(filter(Restrict(witness, 1..M), c))
                BY <7>5, <7>8, filter_range DEF Restrict
              <8>. QED  BY <8>1, <7>4, filter_range DEF History_type
            <7>10. /\ j \in DOMAIN history'
                   /\ j \in 1 .. Len(history')
                   /\ msg = instr(history'[j].st[c])
                   /\ msg.late < time'
              BY <2>1, <7>1, <5>1, <7>3, <7>9, Z3 DEF History_type
            <7>11. msg \in messages'
              BY <7>10, <7>3, <7>msg DEF Inv_messages, sent
            <7>12. messages' = messages
              BY <2>1, <5>1, <5>2, exec_msgs DEF State_type
            <7>. QED  BY <7>1, <7>11, <7>12 DEF msgReceived
          <6>2. ASSUME msgReceived(messages, wit_ins)
                PROVE  msgReceived(witness[N].msgs, wit_ins)
            <7>1. PICK msg \in messages : /\ wit_ins.ins.msg = msg.ins.msg
                                          /\ visible(msg, wit_ins)
                                          /\ msg.late < wit_ins.early
              BY <6>2, visible_cond, instr_type DEF msgReceived, State_type, SystemState, Message, Stream
            <7>2. DOMAIN history = 1 .. Len(history)
              BY History_type, LenProperties
            <7>3. PICK c \in Agent : msg \in sent(Range(project_agent(history,c)))
              BY <7>1, <7>2 DEF Inv_messages, sent, Range, project_agent
            <7>4. Range(project_agent(history, c)) = Range(filter(history, c))
              BY unstutter_range, project_agent_type DEF History_type, filter
            <7>5. Range(filter(history,c)) \subseteq Range(filter(Restrict(witness, 1..U), c))
              BY <3>1, <3>5, myRangeIsPrefix
            <7>6. Range(filter(history,c)) \subseteq { Restrict(witness, 1..U)[i].st[c] : i \in 1 .. U }
              BY <7>5, <3>4, filter_range DEF GoodSystemState
            <7>7. PICK j \in 1 .. U : msg = instr(witness[j].st[c]) /\ msg.ins.type = "send"
              BY <7>3, <7>4, <7>6 DEF sent, Restrict, Message
            <7>8. msg.late < witness[N+1].t
              BY <7>1, <3>2, instr_type, Z3
                 DEF executes, State_type, SystemState, Message, DatedInstruction, GoodNextTime, Stream
            <7>9. msg \in witness[N+1].msgs
              BY <7>7, <7>8, witness_sends_messages DEF Stream
            <7>10. witness[N+1].msgs = witness[N].msgs
              BY <3>2, <5>2, exec_msgs DEF executes, SystemState, Stream
            <7>. QED  BY <7>9, <7>10, <7>1 DEF msgReceived
          <6>. QED  BY <6>1, <6>2
        <5>. /\ witness[N] \in SystemState
             /\ witness[N+1] \in SystemState
             /\ history \in Seq(SystemState)
          BY DEF History_type, Stream
        <5>3. witness[N+1].st[a] = [st |-> state', msgs |-> messages', t |-> time'].st[a]  \* state'[a]
          BY <3>2, <2>1, <5>1, <5>2, exec_from_same_state DEF State_type, executes
        <5>4. Restrict(witness, 1 .. N+1) = Append(Restrict(witness, 1..N), witness[N+1])
          BY DEF Restrict
        <5>5. filter(Append(history, [st |-> state', msgs |-> messages', t |-> time']), a)
            = filter(Append(Restrict(witness, 1..N), witness[N+1]), a)
          BY filter_append, <5>3, <3>2, <3>4
        <5>. QED
          BY history' = Append(history, [st |-> state', msgs |-> messages', t |-> time']), <5>4, <5>5, Zenon
      <4>2. CASE b # a
        <5>1. Last(history).st[b] = state'[b]
          BY <2>1, <4>2 DEF History_type, State_type, Inv
        <5>2. filter(history', b) = filter(history,b)
          BY <2>1, filter_stutter, <5>1, Zenon DEF History_type
        <5>. QED  BY <5>1, <5>2, <3>1, <3>5, IsPrefixPartialOrder, Z3
(* more detailed proof of the QED:
        <5>3 /\ filter(history,b) \in Seq(State)
             /\ filter(Restrict(witness, 1 .. U), b) \in Seq(State)
             /\ filter(Restrict(witness, 1 .. newU), b) \in Seq(State)
             /\ IsPrefix(filter(history, b), filter(Restrict(witness, 1..U), b))
             /\ IsPrefix( filter(Restrict(witness, 1 .. U), b),
                          filter(Restrict(witness, 1 .. newU), b))
           BY <3>1, <3>5 \* Instantation to b.
        <5> QED BY ONLY <5>2, <5>3, IsPrefixPartialOrder, Zenon
*)
      <4>. QED  BY <4>1, <4>2
    <3>. QED  BY <3>3, <3>6 DEF Det
  <2>2. CASE UNCHANGED vars  BY <2>2 DEF Det, vars
  <2>. QED  BY <2>1, <2>2 DEF Next
<1>. QED  BY <1>1, <1>2, Spec_type, Spec_Inv, Spec_inv_messages, PTL DEF Spec 

===========================================================================================
\* Modification History
\* Last modified Wed Jan 27 09:04:16 CET 2016 by merz
\* Last modified Wed Jul 22 15:07:00 CEST 2015 by matthieu
\* Last modified Wed Jul 15 15:27:44 CEST 2015 by merz
